;; Generated package description from sm-c-mode.el  -*- no-byte-compile: t -*-
(define-package "sm-c-mode" "1.1" "C major mode based on SMIE" 'nil :commit "3ef465d3ebed319fdaa3bd650bc19eb11c786fe5" :url "https://elpa.gnu.org/packages/sm-c-mode.html" :authors '(("Stefan Monnier" . "monnier@iro.umontreal.ca")) :maintainer '("Stefan Monnier" . "monnier@iro.umontreal.ca"))
